// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Esizeit.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to find the sky area of a set of exposure images.
//              Overlapping images are handled by removing duplicate pixels.
//              This task is particularly useful for finding the size of 
//              slews.
//
//#include <ctime>
#include <unistd.h> // For getpid
#include<iostream>
#include<fstream>

#include "Tracer.h"
#include "errstr.h"
#include "param.h"

#include "CalServer.h"
#include "Dal.h"
#include "DssBuilder.h"

//#include "ImageRegion.h"
#include "CalSetter.h"
//#include "STime.h"
#include "SscWcsConvertor.h"
#include "Esizeit.h"

// Declare global variable
double pixsize = 3600.0/4.1;

Esizeit::Esizeit()
{
  /* Get the parameters from this command line */
    string imagelist = stringParameter("imagelist");

  // Open the textfile imagelist
    ifstream myReadFile;
    myReadFile.open(imagelist.c_str());
    //char expfile[100];
    string expfile;

  /* Attempt to set the CAL state using the input image header */
  // Not sure this is necessary - probbaly is
    myReadFile >> expfile;

    CalSetter * _calset = new CalSetter();
    _calset->fromImage(expfile);
    _calset->setAccuracyLevel(EXTENDED);

  /* Loop over each image in the list */
    while (!myReadFile.eof()) {

      // Loop over each pixel, quantise and add onto megapix list 
       handleExposure(expfile);

      // Get the next filename
       myReadFile >> expfile;

    } // End of exposure map list

  // Close exposure map list
    myReadFile.close();

  // Remove duplicates from the exposure list
    sort( _megapix.begin(), _megapix.end() );
    _megapix.erase( unique(_megapix.begin(),_megapix.end() ),_megapix.end() );

  // Show the result
    cout<<imagelist<<" : "<<_megapix.size()<<endl;
//    double pixsize = 3600.0/4.1;
    cout<<imagelist<<" : "<<_megapix.size()/pixsize/pixsize<<" (DEG2)"<<endl;
}

Esizeit::~Esizeit()
{
}

void Esizeit::handleExposure(string & expfile)
{

    //double pi180 = 3.1415926535/180.0;

  /* Set the ssclib wcs and attitude structure */
    SscWcsConvertor::setWcs(expfile);
    SscWcsConvertor::setWcsRef(expfile);
    SscWcsConvertor::setAtt(expfile);

  // Open file and read the exposure into a 2-d array
    cout<<"Opening: "<<expfile<<endl;

    ImageData expdata = readImage(expfile);

  // Set the pixel size = 4.1"
 //   double pixsize = 3600.0/4.1;

  // Iterate over each pixel
    while (! expdata.image->isDone()) {

   // Read the next pixel
       PixelElement pixel = expdata.image->currentItem();

   // Get the RA/DEC of this pixel if exposure higher than 1 second
       if (pixel.value()>0.0) {

        // Find the RA/DEC grid position of this pixel
        // We chop the sky into 4.1" x 4.1" pixels
          CoordPair c = SscWcsConvertor::imagePixelsToRaDec(pixel.xIndex(),pixel.yIndex());
          int intRa = int(c.x()/ DEG_TO_RAD * cos(c.y()) * pixsize),
              intDec = int(c.y()/ DEG_TO_RAD * pixsize);

        // Add this into the position vector
          pair<int,int> p = make_pair(intRa, intDec);
          _megapix.push_back(p);
       } else { 
//          cout<<"Pix: "<<pixel.xIndex()<<" "<<pixel.yIndex()<<" "<<pixel.xCentre()<<" "<<pixel.yCentre()<<" "<<pixel.value()<<endl;
       }

       expdata.image->next();
    }

   // Close image file
    dataSetServer->close(expdata.imageset);
}

Esizeit::ImageData Esizeit::readImage(const string & imagefile)
{
 /* Open the input image as a dataset */
    ImageData im;

    im.imageset = dataSetServer->open( imagefile, DataSet::Read );

   /* Make a pixel iterator from the input image */
    if (imagefile.find(":") != string::npos) {
       im.image = new ArrayPixelIterator(imagefile);
    } else {
       im.image = new ArrayPixelIterator(imagefile+":");
    }
   // Initialise the reading of this array
    im.image->first();
   
   /* Return */
    return im;
}
