// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Eupper.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to calculate the count rate upper limit
//              at a given celestial position within an image
//
//              For multiple images the task can either add the exposure times
//              to find a deeper upper limit from a series of observations or
//              average them to extend an energy range, e.g. add XMM pointed
//              images 4000 and 5000 to get the 2-12 keV range
//
//#include <ctime>
#include <iostream>
#include <fstream>
//gcc4.8.2 portability
#include <unistd.h>


#include "Tracer.h"
#include "errstr.h"
#include "param.h"

#include "Dal.h"
#include "Eupper.h"
#include "UpperLimitSolver.h"
#include "ArrayPixelIterator.h"
#include "SscWcsConvertor.h"

/* Method: Eupper() - Constructor */
Eupper::Eupper()
{
  PROC;

    //cout<<"CLOCK: Starting: "<<clock();
  /* Get the parameters from this command line */
    string test = stringParameter("imageset");
    vector<string> imagefiles = splitString(test);
    vector<string> exposurefiles = splitString(stringParameter("exposuremap"));
    vector<string> bckgndfiles = splitString(stringParameter("bkgimageset"));
    double src_ra = realParameter("srcra");
    double src_dec = realParameter("srcdec");
    double src_rad = realParameter("srcradius");

  // Set background region 
    double bck_ra = src_ra, bck_dec = src_dec, bck_rad_in = src_rad;
 
    if (realParameter("bgdra")) 
       bck_ra = realParameter("bgdra");
    if (realParameter("bgddec"))
       bck_dec = realParameter("bgddec");
    if (realParameter("bgdrinner"))
       bck_rad_in = realParameter("bgdrinner");

    double bck_rad_out = realParameter("bgdrouter");

    double ulsig = realParameter("ulsig");

    //cout<<"Bck: "<<bck_ra<<" "<<bck_dec<<" "<<bck_rad_in<<" "<<bck_rad_out<<endl;

  /* Create variables to hold cumulative counts and exposure */
    double src_counts=0.0, bck_counts=0.0, toterr=0.0, exposure=0.0, eef=0.0,
           area_ratio=0.0; 
    //string obsids=" ",date_min="4000-01-01T00:00:00",date_max="0000-01-01T00:00:00";
    string obsids=" ",date_min="Not Set",date_max="----------";

  /* Loop over each image */
    for (unsigned i=0; i<imagefiles.size(); i++)
    {
      /* Find exposure at the source centre from exposure map if
         present. Skip image if exposure=0 */
       double expbit = getExposure(exposurefiles[i], src_ra,
                                    src_dec, 1.0);
       if (expbit>0) {

         /* Get a pixel iterator for the image */
          ImageData srcImage = readImage(imagefiles[i]);
  
         /* Add on the exposure for this image */
          if (exposurefiles[i] != "NotSet") 
             exposure += expbit;
          else
             exposure += srcImage.exposure;

         /* Calculate the counts in the source region */
          COUNTS source = counts(srcImage, src_ra, src_dec, 0.0, src_rad);
          src_counts += source.counts;
    
         /* Process the background information: If a background map has been 
         supplied use this in conjunciton with the source spatial expression,
         If a map hasn´t been set use the source image with the bckgnd region */
          COUNTS bck;
          if (bckgndfiles[0] != "NotSet") {
             ImageData bgdImage = readImage(bckgndfiles[i]);
             bck = counts(bgdImage, src_ra, src_dec, 0.0, src_rad);
          } else {
             bck = counts(srcImage, bck_ra, bck_dec, bck_rad_in, bck_rad_out);
          }

         /* Find ratio of source to background region areas and normalise background */
          area_ratio = source.area / bck.area;
          bck_counts += bck.counts * area_ratio;

         /* Find encircled energy factor for source region 
            Implicitly assumes that eef is the same for each image */
          eef = realParameter("eef");  
          if (! booleanParameter("witheef"))   // calculate if not set as paremeter
             eef = getEEF(srcImage, src_rad);

      /* Calculate the error on the bckgnd subtracted count rate */
      // ? Is it correct to just add here - should revisit the non-bayesian calc //
          toterr += sqrt(source.counts + bck.counts
                                       * area_ratio*area_ratio) / eef;

       //cout<<"Toterr: "<<toterr<<" "<<source.counts<<" "<<bck.counts<<" "<<area_ratio<<" "<<eef<<endl;
      /* Produce strings for obsid(s) and min/max observation dates*/
          obsids = obsids + " " + srcImage.obsid;
          date_min=min(date_min,srcImage.start_date);
          date_max=max(date_max,srcImage.end_date);
       }
    }

  /* Use the average exposure if required. This option would typically be used 
     for joining images from different energy bands */
    if (stringParameter("expmode") == "average")
           exposure = exposure / imagefiles.size();

  /* Calculate upper limit */
    double upplim = calcUpperlimit(exposure, src_counts,
                              bck_counts, toterr, ulsig, eef);

  /* Output useful debug if required */
    //errstr<< "Source/back counts: "<<source.counts<<" "<<bck.counts<<endl
    errstr<< "Source/back counts: "<<src_counts<<" "<<bck_counts<<endl
        << "Area ratio: "<<area_ratio<<endl
        << "Exposure: "<<exposure<<endl
        << "EEF: "<<eef
        << Sas::message(Msg::App, Msg::Verbose);

  /* Calculate the bckgnd subtracted count rate and error */
    double bkg_sub_crate = (src_counts - bck_counts) / eef / exposure;
    double bkg_sub_crate_err = toterr / exposure;

  /* Output the result */
    ostringstream outs;
    outs.precision(8);

    if (booleanParameter("details")) {
       outs<<"Source/back counts: "<<src_counts<<" "<<bck_counts<<endl
           <<"Area ratio: "<<area_ratio<<endl
           << "Exposure: "<<exposure<<endl
           << "EEF: "<<eef<<endl
           << "Obsid: "<<obsids<<endl
           << "Dates: "<<date_min<<" "<<date_max<<endl
           << "Bckgnd subtracted source c/r: "<<bkg_sub_crate
           << " +/- "<<bkg_sub_crate_err<<" c/s" <<endl;
    }
    outs<<" Statistical upper limit c/r: "<<upplim<<" c/s" <<endl;

    if (booleanParameter("withoutputfile")) {

      /* Open a stream to an output text file */
       ofstream outfile;
       string textfile=stringParameter("output");
       outfile.open(textfile.c_str());
       outfile<<outs.str();
       outfile.close();
    
    } else { // Write to standard error

       errstr<<outs.str()
             <<Sas::message(Msg::App, Msg::Always);
    }
  // } REMOVE

  /* Set output parameters */

  ENDPROC;
}

/* Destructor */
Eupper:: ~Eupper()
{
  PROC;

    // CLOSE EVERYTHING DOWN PROPERLY
    //delete srcreg;            srcreg=0;

  ENDPROC;
};

/* Eupper::calcUpplim - calculates the upper limit for total counts
,n, background counts, b and confidence limit, cl */
double Eupper::calcUpplim(double n, double b, double cl)
{
   UpperLimitSolver u = UpperLimitSolver(static_cast<long int>(n),b);
   std::pair<double, double> limits = u.bayesian(cl);
   //double lower = limits.first, upper=limits.second;
   double upper=limits.second;

   //cout<<"LIMITS: "<<lower<<" "<<upper<<endl;

   RETURN upper;
}

Eupper::ImageData Eupper::readImage(const string & imagefile)
{
 /* Open the input image as a dataset */
    ImageData im;

    im.imageset = dataSetServer->open( imagefile, DataSet::Read );

   // Set up the WCS matrix - this is used internally within SSCLIB
   // conversion routines
    SscWcsConvertor::setWcs(imagefile);

  /* Check whether image axes are inverted */
 //   bool invert_axes = checkaxes(array);

  /* Make a pixel iterator from the input image */
  /* SHOULD handle inverted axes HERE !! */
    if (imagefile.find(":") != string::npos) {
       im.image = new ArrayPixelIterator(imagefile);
    } else {
       im.image = new ArrayPixelIterator(imagefile+":");
    }
    //cout<<"CLOCK: got the array"<<clock()<<endl;

  // Initialise the reading of this array
    im.image->first();

  /* Find the instrument and observing mode */
    if (im.imageset->hasAttribute("INSTRUME")) 
       im.instrument = im.imageset->attribute("INSTRUME")->asString();
    else
       errstr<<"Attribute INSTRUME not found in image header"
                   << Sas::warning("InstrumentNotFound");

    if (im.imageset->hasAttribute("OBS_MODE")) 
       im.obsmode = im.imageset->attribute("OBS_MODE")->asString();
    else
       errstr<<"Attribute OBS_MODE not found in image header"
                   << Sas::warning("ObsModeNotFound");

  /* Attempt to get the observation id from the main array */
    if (im.imageset->hasAttribute("OBS_ID"))
       im.obsid = im.imageset->attribute("OBS_ID")->asString();
    else
       im.obsid = "----------";

  /* Attempt to get the observation dates from the main array */
    if (im.imageset->hasAttribute("DATE-OBS")) 
       im.start_date = im.imageset->attribute("DATE-OBS")->asString();
    else if (im.imageset->hasAttribute("DATE_OBS"))
       im.start_date = im.imageset->attribute("DATE_OBS")->asString();
    else
       im.start_date = "----------";

    if (im.imageset->hasAttribute("DATE-END"))
       im.end_date = im.imageset->attribute("DATE-END")->asString();
    else if (im.imageset->hasAttribute("DATE_END"))
       im.end_date = im.imageset->attribute("DATE_END")->asString();
    else
       im.end_date = "----------";

  /* Get the exposure from the main array */
    Array * array;
    if (imagefile.find(":") != string::npos) {
       array = im.imageset->array(imagefile);
    } else {
       array = im.imageset->array(imagefile+":");
    }

    if (array->hasAttribute("EXPOSURE")) {
       im.exposure = array->attribute("EXPOSURE")->asReal();
    } else {
       im.exposure = 1.0;
    }

   // If this is a detector image then we also need to set the 
   // spacraft and instrument attitude matrices
    if (im.imageset->hasAttribute("CTYPE1") ) {
       if (im.imageset->attribute("CTYPE1")->asString() == "DETX") {
          SscWcsConvertor::setAtt(imagefile); } }
    else if (array->hasAttribute("CTYPE1") ) {
       if (array->attribute("CTYPE1")->asString() == "DETX" ) {
          SscWcsConvertor::setAtt(imagefile); }}
 
  /* Return */
    return im;
}


/* Method - counts() - returns the total counts in the image region */
Eupper::COUNTS Eupper::counts(ImageData imdat, double ra, double dec,
                      double rad_inner, double rad_outer)
{
  PROC;

     Eupper::COUNTS result;

    // Find image pixel index which contains this RA, DEC
     PixelElement pixel = findPixel(imdat, ra, dec);

     int xcen = pixel.xIndex(), ycen = pixel.yIndex();

    // Calculate width of region in pixels
    // Ensure pixel width is in units of degrees
     double convtodeg=1.0;
     if (imdat.image->wcsconvertor()->xType().find("DET") != string::npos)
        convtodeg=72000.0;
 
     double xhwid = rad_outer / (abs(pixel.xWidth()) / convtodeg);
     double yhwid = rad_outer / (pixel.yWidth() / convtodeg);

     //cout<<"Pixel index: "<<xcen<<" "<<ycen<<" "<<xhwid
       //  <<" "<<yhwid<<endl;

     double rout_pix = xhwid, 
            rinn_pix = rad_inner / (abs(pixel.xWidth()) / convtodeg);

    // Find X and Y ranges of pixels
     int ixmin=pixel.xIndex() - INT(xhwid) - 1;
     int ixmax=pixel.xIndex() + INT(xhwid) + 1;
     int iymin=pixel.yIndex() - INT(yhwid) - 1;
     int iymax=pixel.yIndex() + INT(yhwid) + 1; 

     int xbins = imdat.image->xbins();
     int ybins = imdat.image->ybins();

     ixmin = min( max(1,ixmin), (xbins));
     iymin = min( max(1,iymin), (ybins));
     ixmax = min( max(1,ixmax), (xbins));
     iymax = min( max(1,iymax), (ybins));

     double total=0.0, offset=0.0;
     int npix_in_area=0;

     //cout<<"CLOCK:Iterating "<<clock()<<endl;
     //cout<<"Iterating over pixels: "<<ixmin<<" "<<ixmax<<" "<<iymin<<" "<<iymax<<endl;

   // Loop over the X,Y range which contains the region
     for (int iy = iymin; iy<=iymax; iy++)
     {
        for (int ix = ixmin; ix<=ixmax; ix++)
        {
        //   imdat.image->setIndex( (iy-1) * xbins + ix -1 );

          // Check if this pixel is within annulus
           offset = sqrt(float( ( xcen - ix )*( xcen - ix )
                   + ( ycen - iy )*( ycen - iy ) ));
           
           if (offset <= rout_pix && offset>= rinn_pix) {
              //total = total+pixel.value();
              unsigned long index = (iy-1) * xbins + ix -1;
              total = total+imdat.image->valueAt(index);
              npix_in_area++;
           //cout<<"DEBUG: "<<imdat.image->valueAt(index)<<" "<<ix<<" "<<iy<<endl;
           }
           
        }
     }
     result.counts = total;
     result.area = double(npix_in_area);

     //cout<<"CLOCK:Stopped Iterating "<<clock();
     cout<<"Counts are: "<<result.counts<<" "<<result.area<<endl;

  RETURN result;
}

double Eupper::getExposure(const string &exposurefile, double src_ra,
                              double src_dec, double defaultExp) 
{   
  PROC;

    double exposure = 0.0;
    if (exposurefile != "NotSet") {

      // Open the exposure file and get the data array
       ImageData expdata = readImage(exposurefile);

      // Get the pixel which contains the source centre
       PixelElement pixel = findPixel(expdata, src_ra, src_dec);

       exposure = pixel.value();

    } else {
       exposure = defaultExp;
    }

    return exposure;
}

double Eupper::calcUpperlimit(double exposure, double src_counts, 
                              double bck_counts, double toterr,
                              double cl, double eef) 
{
  // bck_counts is the background rate normalised to the source region area

   double upplim = 0.0;

   if (exposure > 0.0) {  // Leave upper limit as 0 if exposure <= 0
     if (src_counts < 80.0) { // Use Bayesian statistics
     //if (src_counts < 800.0) { // Use Bayesian statistics
        upplim = calcUpplim(src_counts, bck_counts, cl);
     } else {// Use source counts + 2 x error

    /* Need to convert between confidence interval and sigma.
       Can't find the equation right now - so hardcode for 1,2 and 3-sigma */
       double fac=0.0;
       if ( cl>0.67 && cl<0.69)
          fac=1.0;
       else if ( cl>0.98)
          fac=3.0;
       else
          fac=2.0;

       //upplim = ((src_counts - bck_counts) + fac * ( toterr * eef ));
       upplim = (max((src_counts - bck_counts),0.0) + fac * ( toterr * eef ));
   //   cout<<"DBG: "<<toterr<<" "<<eef<<" "<<exposure<<" "<<fac<<" "<<upplim<<endl;
     //  cout<<"Comp: "<<upplim<<" "<<calcUpplim(src_counts, bck_counts, cl)<<endl;
     }
   }

 //  cout<<"Test 0 3: "<<calcUpplim(0.0, 3.0, 0.954)<<endl;
//   cout<<"Test 3 0: "<<calcUpplim(3.0, 0.0, 0.954)<<endl;
 //  cout<<"Test 5 5: "<<calcUpplim(5.0, 5.0, 0.954)<<endl;
  // cout<<"Test 10 10: "<<calcUpplim(10.0, 10.0, 0.954)<<endl;
 //  cout<<"Test 50 50: "<<calcUpplim(50.0, 50.0, 0.954)<<endl;
 //  cout<<"Test 100 100: "<<calcUpplim(100.0, 100.0, 0.954)<<endl;
 //  cout<<"Test 1000 1000: "<<calcUpplim(1000.0, 1000.0, 0.954)<<endl;

  // Return the upper limit in c/s corrected for the encircled energy factor
   return upplim / eef / exposure;
}

double Eupper::getEEF(ImageData srcImage, double src_rad_deg) const
{
  // Hardwire these correction factors to allow the code to run on any image
  // and for any instrument
    double tol=1.0; // arcsecs
    double eef=1.0; // full coverage by default
    double src_rad = src_rad_deg * 3600.0; // In arcseconds
 
  // EPIC - values are for EPN
    if (srcImage.instrument.substr(0,3) == "EPN" || 
        srcImage.instrument.substr(0,4) == "EMOS" ) {

       if (srcImage.obsmode.substr(0,4) == "SLEW") {
          if ( abs(src_rad - 30.0) < tol ) { 
             eef=0.845;  // This is the slew PSF for 1.5 keV - not very E-dept.
          } else if ( abs(src_rad - 45.0) < tol ) {
             eef=0.907;
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.936;
          }
       } else { // POINTING we assume - set for 6' off-axis, otherwise slow.  
          if ( abs(src_rad - 30.0) < tol ) { 
             eef=0.876;  // This is the pointed PSF for 1.5 keV - not very E-dep
          } else if ( abs(src_rad - 45.0) < tol ) {
             eef=0.927;
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.951;
          }
       }
    // ROSAT: the PSF is energy-independent from 100-200 arcsecs radius (60-90%)
    // in the RASS.
    } else if (srcImage.instrument.substr(0,4) == "PSPC") { // ROSAT
       if (srcImage.obsmode.substr(0,6) == "SURVEY") {
          if ( abs(src_rad - 30.0) < tol ) { 
             eef=0.2;  // This is TBD
          } else if ( abs(src_rad - 45.0) < tol ) {
             eef=0.35;
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.4;
          } else if ( abs(src_rad - 100.0) < tol ) {
             eef=0.58;
          }
       } else { // Pointed data - VERY position-dependent
             errstr<<"No EEF value for ROSAT POINTED data"
                   << Sas::warning("DataNotSupported");
       }
    }
    else {
       errstr<<"Instrument / OBS_MODE combination not recognised, using "
             <<"encircled energy fraction = 1.0"
             << Sas::warning("DataNotSupported");
    }
    return eef;
}

bool Eupper::checkaxes(Array *array) 
{
  PROC;

   /* If the X axis name is "Y" or "DETY" or "DEC" then the image
      is inverted */
    if ( array->attribute("CTYPE1")->asString().find("Y") != string::npos ||
         array->attribute("CTYPE1")->asString().find("DEC") != string::npos )
    {
       ALTRETURNV( true );
    }
    RETURN false;
}

// findPixel - a method to find the pixel which contains a given ra, dec
PixelElement Eupper::findPixel(Eupper::ImageData imdat, double ra, 
                                                   double dec) 
{
   // Input ra, dec should be in degrees - here we convert to radians
   double raR=ra*DEG_TO_RAD, decR=dec*DEG_TO_RAD;

   CoordPair p = SscWcsConvertor::raDecToImagePixels(raR, decR);

   int ix = floor(p.x()+0.5), iy=floor(p.y()+0.5);

   imdat.image->setIndex( (iy-1) * imdat.image->xbins() + ix -1 );
   PixelElement pixel = imdat.image->currentItem();

   return pixel;
}

// Split string into words
vector<string> Eupper::splitString(const string &text) {
  vector<string> tokens;
  char sep=' ';
  size_t start = 0, end = 0;
  end = text.find(sep, start);
  while (end != std::string::npos) {
    tokens.push_back(text.substr(start, end - start));
    start = end + 1;
    end = text.find(sep, start);
  }
  tokens.push_back(text.substr(start));
  return tokens;
}
