// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// etimeget.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Metatask which runs evselect, arfgen and rmfgen to 
//              produce source and background timetra and the appropriate
//              ARF and RMF.
//
//#include <set>          // C++
//#include <algorithm>
//#include <strstream.h>
//#include <unistd.h>
#include <Tracer.h>
#include <metatask.h>
#include <errstr.h>
#include <param.h>
#include "Evaluator.h"
#include "DataFilter.h"   // tasks

#include "etimeget.h"
#include "DatafileAttributes.h"
#include "LightCurveMaker.h"

void etimeget()
{
  PROC;

  /* Set the source, backgnd and arf filenames */
    string sourceset, backset, arfset, outrmfset;

    if  (booleanParameter("withfilestem")) {
        string filestem = stringParameter("filestem");
        sourceset = filestem+"_srcrate.lc";
        backset   = filestem+"_bgdrate.lc";
    } else {
        sourceset = stringParameter("srcrateset");
        backset   = stringParameter("bckrateset");
    }

  /* Get the parameters from this command line */
    string srcexp = stringParameter("srcexp");
    string backexp = stringParameter("backexp");
    string eventfile = stringParameter("table");
    string timebinstyle = stringParameter("timebinstyle");
    double timebin = realParameter("timebinwidth");

  /* Read observation attributes from the event file */
    DatafileAttributes * timeAttributes = new DatafileAttributes(eventfile);

  /* Check that the input expressions are valid */
    // First expand any macros 
    DataSet *set = dataSetServer->open(timeAttributes->datafile(), Dal::Read);
    Table *table = set->table(0);
    srcexp = DataFilter::expandSelectionExpression(srcexp,*table);
    backexp = DataFilter::expandSelectionExpression(backexp,*table);
    dataSetServer->close(set);
    
    if (! Evaluator::isValidExpression(srcexp)) {
       errstr<<"Invalid expression: "<<srcexp
             << Sas::error("InvalidExpression");
    }
    else if (! Evaluator::isValidExpression(backexp)) {
       errstr<<"Invalid expression: "<<backexp
             << Sas::error("InvalidExpression");
    }

  /*----------------------------------------------------------
     Set the timebin size for the lightcurve
  ----------------------------------------------------------*/

   /* If timebinstyle=intrinsic, find the intrinsic bin width for
      this instrument and observing mode */

    if (timebinstyle == "intrinsic") {
       timebin = timeAttributes->intrinsicFrameTime();
       errstr<<"Intrinsic frame time is "<<timebin
          << Sas::message(Msg::App, Msg::Sparse);

    }
    else if (timebinstyle == "parameter") 
    {
       // timebin set earlier
    }
    else {   // Bin style not recognised
       errstr<<"Invalid time bin style: "<<timebinstyle
             << Sas::error("InvalidTimebinstyle");
    }

  /* Create the source time series */
    errstr<<"Creating source time series..."
          << Sas::message(Msg::App, Msg::Sparse);

    LightCurveMaker stime = LightCurveMaker(sourceset, timeAttributes,
                                                   srcexp, timebin);

  /* Create the background timetrum */
    errstr<<"Creating background time series..."
          << Sas::message(Msg::App, Msg::Sparse);

    LightCurveMaker btime = LightCurveMaker(backset, timeAttributes,
                                                  backexp, timebin);
  /* Shutdown */
    delete timeAttributes; timeAttributes=0;

    closeParameters();

  ENDPROC;
}
